use Class::DBI;

package Class::DBI;

{
	no warnings;
	no strict;
	sub _do_search {
		my ($proto, $search_type, @args) = @_;
		my $class = ref $proto || $proto;
		
		@args = %{ $args[0] } if ref $args[0] eq "HASH";

		my (@cols, @vals);
		my $search_opts = @args % 2 ? pop @args : {};

		$search_opts->{offset} = int($search_opts->{page}) * int($search_opts->{page_size})  if ($search_opts->{page_size});
		$search_opts->{_placeholder} ||= '?';

		while (my ($col, $val) = splice @args, 0, 2) {
			my $column = $class->find_column($col)
				|| (List::Util::first { $_->accessor eq $col } $class->columns)
				|| $class->_croak("$col is not a column of $class");

			push @cols, $column;
			push @vals, $class->_deflated_column($column, $val);
		}

		my $frag = join " AND ",
		map defined($vals[$_]) ? "$cols[$_] $search_type $$search_opts{_placeholder}" : "$cols[$_] IS NULL",
			0 .. $#cols;

		$frag .= " ORDER BY $search_opts->{order_by}"
			if $search_opts->{order_by};
		$frag .= " LIMIT $search_opts->{limit}"
			if $search_opts->{limit};
		$frag .= " OFFSET $search_opts->{offset}"
			if ($search_opts->{limit} && defined($search_opts->{offset}));

		return $class->sth_to_objects($class->sql_Retrieve($frag),
			[ grep defined, @vals ]);
	}
}

sub search_fti {
	my $self = shift;
	my @args = @_;
	if (ref($args[-1]) eq 'HASH') {
		$args[-1]->{_placeholder} = "to_tsquery('default',?)";
	} else {
		push @args, {_placeholder => "to_tsquery('default',?)"};
	}
	$self->_do_search("@@"  => @args);
}



#-------------------------------------------------------------------------------
package OpenILS::App::Storage;
use base qw/OpenILS::Application/;
use OpenILS::Utils::Config;
use OpenILS::DOM;
use OpenILS::AppSession;
use OpenILS::EX qw/:try/;
use OpenILS::DomainObject::oilsResponse;
use OpenILS::DomainObject::oilsPrimitive;
use OpenILS::Utils::Logger qw/:level/;

my $log = "OpenILS::Utils::Logger";

sub DESTROY {};

our $_db_driver;
our %_db_params;
our %_METHODS;

sub run {
	my $self = shift;
	my @args = @_;
	if (defined $self->{params}) {
		if ($self->{params} > scalar(@args)) {
			throw OpenILS::EX::InvalidArg ("Not enough arguments to $$self{name}: ".join(', ',@args));
		}
	}
	return $self->SUPER::run(@args);
}

sub method_lookup {
	my $self = shift;
	my $method = shift;
	my $proto = shift;

	my $super_lookup = $self->SUPER::method_lookup($method,$proto);
	return $super_lookup if (ref $super_lookup);

	my $class = ref($self) || $self;

	$log->debug("Specialized lookup of [$method] in [$class]", INTERNAL);
	
	if (exists $_METHODS{$method}) {
		return bless $_METHODS{$method} => $class;
	}
	return undef;
}

sub is_action {
	my $self = shift;
	if (my $can = $self->can($self->{name} . '_action')) {
		return $can->();
	}
	return $self->{action};
}



sub initialize {
	return $_db_driver if (defined $_db_driver);
	my $conf = OpenILS::Utils::Config->current;

	$log->debug('Initializing ' . __PACKAGE__ . '...', DEBUG);

	my $driver = $conf->storage->database->driver;
	$_db_params{driver} = $conf->storage->database->driver;
	$_db_params{user} = $conf->storage->database->user;
	$_db_params{pw} = $conf->storage->database->pw;
	$_db_params{host} = $conf->storage->database->host;
	$_db_params{database} = $conf->storage->database->database;

	$_db_driver = "OpenILS::App::Storage::$driver";

	$log->debug("** DB Params for $_db_driver :\n\t" . join("\n\t", map { "$_ ==> $_db_params{$_}" } keys %_db_params), INTERNAL);

	push @OpenILS::App::Storage::CDBI::ISA, $_db_driver;

	eval "use $_db_driver;";
	throw OpenILS::EX::Config ( "Can't load $_db_driver!  :  $@" ) if ($@);

	%_METHODS = (
		'storage.opac.MODS_from_metarecord' => {
			code	=> \&{$_db_driver.'::storage_opac_MODS_from_metarecord'},
			name	=> $_db_driver.'::storage_opac_MODS_from_metarecord',
			action	=> 1,
			argc	=> 1,
		},
		'storage.opac.hitlist.count' => {
			code	=> \&{$_db_driver.'::storage_opac_hitlist_count'},
			name	=> 'storage.opac.hitlist.count',
			action	=> 1,
			argc	=> 2,
		},
		'storage.opac.metarecord_copy_count' => {
			code	=> \&{$_db_driver.'::storage_opac_metarecord_copy_count'},
			name	=> $_db_driver.'::storage_opac_metarecord_copy_count',
			action	=> 1,
			argc	=> 1,
		},
		'storage.opac.record_copy_count' => {
			code	=> \&{$_db_driver.'::storage_opac_record_copy_count'},
			name	=> $_db_driver.'::storage_opac_record_copy_count',
			action	=> 1,
			argc	=> 1,
		},
		'storage.opac.metarecord_avail_count' => {
			code	=> \&{$_db_driver.'::storage_opac_metarecord_avail_count'},
			name	=> $_db_driver.'::storage_opac_metarecord_avail_count',
			action	=> 1,
			argc	=> 1,
		},
		'storage.opac.record_avail_count' => {
			code	=> \&{$_db_driver.'::storage_opac_record_avail_count'},
			name	=> $_db_driver.'::storage_opac_record_avail_count',
			action	=> 1,
			argc	=> 1,
		},
		'storage.opac.hitlist' => {
			code	=> \&{$_db_driver.'::storage_opac_hitlist'},
			name	=> $_db_driver.'::storage_opac_hitlist',
			action	=> 1,
			argc	=> 2,
		},
		'storage.opac.hitlist.nosort' => {
			code	=> \&{$_db_driver.'::storage_opac_hitlist_nosort'},
			name	=> $_db_driver.'::storage_opac_hitlist_nosort',
			action	=> 1,
			argc	=> 2,
		},
		'storage.opac.recordlist_from_metarecord' => {
			code	=> \&{$_db_driver.'::storage_opac_recordlist_from_metarecord'},
			name	=> $_db_driver.'::storage_opac_recordlist_from_metarecord',
			action	=> 1,
			argc	=> 1,
		},
		'storage.opac.callnumber_list_from_record' => {
			code	=> \&{$_db_driver.'::storage_opac_callnumber_list_from_record'},
			name	=> $_db_driver.'::storage_opac_callnumber_list_from_record',
			action	=> 1,
			argc	=> 1,
		},
		'storage.opac.callnumber_list_from_metarecord' => {
			code	=> \&{$_db_driver.'::storage_opac_callnumber_list_from_metarecord'},
			name	=> $_db_driver.'::storage_opac_callnumber_list_from_metarecord',
			action	=> 1,
			argc	=> 1,
		},
		'storage.general.retrieve_limited' => {
			code	=> \&{$_db_driver.'::storage_general_retrieve_limited'},
			name	=> $_db_driver.'::storage_general_retrieve_limited',
			action	=> 1,
			argc	=> 2,
		},
	);


}

sub child_init {

	$log->debug('Running child_init for ' . __PACKAGE__ . '...', DEBUG);
	
	return 1 if ($_db_driver->db_Main(%_db_params));
	return 0;
}

sub general_search_1_action { 1 }
sub general_search_1 {
	my $session = shift;
	my $search = shift;

	my $response = new OpenILS::DomainObject::oilsResult;

	my $obj = eval { OpenILS::DOM->new->parse_string($search); };
	if( $@ ) {              

		$log->debug( "Received bogus XML", INFO );
		$log->debug( "Bogus XML data: \n $search \n", INTERNAL );
		my $res = OpenILS::DomainObject::oilsXMLParseError->new( status => "XML Parse Error --- $search" );

		$response->content( $res );
		return $response;
	}

}

sub getBiblioFieldMaps_1_action { 1 }
sub getBiblioFieldMaps_1 {
	$client = shift;
	$log->debug(" Executing [getBiblioFieldMaps_1]",INTERNAL);
	my @fields = biblio::metarecord_field_map->retrieve_all;
	$log->debug(" Got ".scalar(@fields)." fields",INTERNAL);
	my $a_o_h = _cdbi_list2oilsArray_of_oilsHash(@fields);

	my $response = new OpenILS::DomainObject::oilsResult;
	$response->content( $a_o_h );
	return $response;
}	

sub getBiblioFieldMapClasses_1_action { 1 }
sub getBiblioFieldMapClasses_1 {
	$client = shift;
	my @classes = biblio::metarecord_field_class_map->retrieve_all;
	my $a_o_h = _cdbi_list2oilsArray_of_oilsHash(@classes);

	my $response = new OpenILS::DomainObject::oilsResult;
	$response->content( $a_o_h );
	return $response;
}	


sub MARC21slimByTcn_1_action { 1 }
sub MARC21slimByTcn_1 {
	my $client = shift;
	my $tcn = shift;

	my ($record) = biblio::record->search( control => $tcn );

	return unless ($record);

	my $marc = $record->datum;

	my $response = new OpenILS::DomainObject::oilsResult;
	$response->content( OpenILS::DomainObject::oilsScalar->new( $marc ) );

	return $response;
}

sub MARC21slimByBarcode_1_action { 1 }
sub MARC21slimByBarcode_1 {
	my $client = shift;
	my $bc = shift;

	$log->debug(" Executing [MARC21slimByBarcode_1] for [$bc]",INTERNAL);

	my ($copy) = asset::copy->search( barcode => $bc );
	my $record = $copy->call_number->record;

	my $marc = $record->datum;

	return unless ($record);

	$log->debug(" Got MARC [$marc] for [$bc]",INTERNAL);

	my $response = new OpenILS::DomainObject::oilsResult;

	$log->debug(" Created a response: ".$response->toString,INTERNAL);

	my $rec = OpenILS::DomainObject::oilsScalar->new( $marc );

	$response->content( $rec );

	$log->debug(" Added content to response: ".$response->toString,INTERNAL);

	return $response;
}

sub MetarecordByFieldClassSearch_1_action { 1 }
sub MetarecordByFieldClassSearch_1 {
	my $client = shift;
	my $field_class = shift;
	my $query = shift;

	$log->debug(" Executing [MetarecordByFieldClassSearch_1] for [$query]",INTERNAL);
	
	my @records = biblio::metarecord_field_entry->field_class_entry_search( $field_class, $query );
	
	return unless (@records);

	$log->debug(" ... Got ".scalar(@records)." rows ...",INTERNAL);

	for my $rec ( @records ) {
		$client->respond( _cdbi2oilsHash($rec) );
	}

	return;
}

sub MARC21slimByFieldClassSearch_1_action { 1 }
sub MARC21slimByFieldClassSearch_1 {
	my $client = shift;
	my $field_class = shift;
	my $query = shift;

	$log->debug(" Executing [MARC21slimByFieldClassSearch_1] for [$query]",INTERNAL);
	
	my @records = biblio::record->field_class_entry_search( $field_class, $query );
	
	return unless (@records);

	$log->debug(" ... Got ".scalar(@records)." rows ...",INTERNAL);

	my $scalar;
	my $obj = tie $scalar, OpenILS::DomainObject::oilsScalar;

	for my $rec ( @records ) {
		try {
			$scalar = $rec->datum;
			$client->respond( $obj );
		} catch Error with {
			my $e = shift;
			$e->{'-text'} .= "\n\n" . $rec->datum;
			throw $e;
		};
	}

	return;
}

sub _cdbi2oilsHash {
	my $obj = shift;
	return new OpenILS::DomainObjectCollection::oilsHash ( map { ( $_ => $obj->$_ ) } $obj->columns );
}

sub _cdbi_list2oilsArray_of_oilsHash {
	my @objs = @_;
	return new OpenILS::DomainObjectCollection::oilsArray ( map { _cdbi2oilsHash($_) } @objs );
}

#-------------------------------------------------------------------------------
package OpenILS::App::Storage::CDBI;
use vars qw/@ISA/;

1;
