package OpenILS::EX;
use Error qw(:try);
use base qw( OpenILS Error );
use OpenILS::Utils::Logger;

my $log = "OpenILS::Utils::Logger";
$Error::Debug = 1;

sub new {
	my( $class, $message ) = @_;
	$class = ref( $class ) || $class;
	my $self = {};
	$self->{'msg'} = ${$class . '::ex_msg_header'} ." \n$message";
	return bless( $self, $class );
}	

sub message() { return $_[0]->{'msg'}; }

sub DESTROY{}


=head1 OpenILS::EX

Top level exception.  This class logs an exception when it is thrown.  Exception subclasses
should subclass one of OpenILS::EX::INFO, NOTICE, WARN, ERROR, CRITICAL, and PANIC and provide
a new() method that takes a message and a message() method that returns that message.

=cut

=head2 Synopsis


	throw OpenILS::EX::Jabber ("I Am Dying");

	OpenILS::EX::InvalidArg->throw( "Another way" );

	my $je = OpenILS::EX::Jabber->new( "I Cannot Connect" );
	$je->throw();


	See OpenILS/EX.pm for example subclasses.

=cut

# Log myself and throw myself

#sub message() { shift->alert_abstract(); }

#sub new() { shift->alert_abstract(); }

sub throw() {

	my $self = shift;

	if( ! ref( $self ) || scalar( @_ ) ) {
		$self = $self->new( @_ );
	}

	if(		$self->class->isa( "OpenILS::EX::INFO" )	||
				$self->class->isa( "OpenILS::EX::NOTICE" ) ||
				$self->class->isa( "OpenILS::EX::WARN" ) ) {

		$log->debug( $self->stringify(), $log->DEBUG );
	}

	else{ $log->debug( $self->stringify(), $log->ERROR ); }
	
	$self->SUPER::throw;
}


sub stringify() {

	my $self = shift;
	my $ctime = localtime();
	my( $package, $file, $line) = get_caller();
	my $name = ref( $self );
	my $msg = $self->message();

	$msg =~ s/^/Mess: /mg;

	return "  * ! EXCEPTION ! * \nTYPE: $name\n$msg\n".
		"Loc.: $line $package \nLoc.: $file \nTime: $ctime\n";
}


# --- determine the originating caller of this exception
sub get_caller() {

	$package = caller();
	my $x = 0;
	while( $package->isa( "Error" ) || $package =~ /^Error::/ ) { 
		$package = caller( ++$x );
	}
	return (caller($x));
}




# -------------------------------------------------------------------
# -------------------------------------------------------------------

# Top level exception subclasses defining the different exception
# levels.

# -------------------------------------------------------------------

package OpenILS::EX::INFO;
use base qw(OpenILS::EX);
our $ex_msg_header = "System INFO";

# -------------------------------------------------------------------

package OpenILS::EX::NOTICE;
use base qw(OpenILS::EX);
our $ex_msg_header = "System NOTICE";

# -------------------------------------------------------------------

package OpenILS::EX::WARN;
use base qw(OpenILS::EX);
our $ex_msg_header = "System WARNING";

# -------------------------------------------------------------------

package OpenILS::EX::ERROR;
use base qw(OpenILS::EX);
our $ex_msg_header = "System ERROR";

# -------------------------------------------------------------------

package OpenILS::EX::CRITICAL;
use base qw(OpenILS::EX);
our $ex_msg_header = "System CRITICAL";

# -------------------------------------------------------------------

package OpenILS::EX::PANIC;
use base qw(OpenILS::EX);
our $ex_msg_header = "System PANIC";

# -------------------------------------------------------------------
# -------------------------------------------------------------------

# Some basic exceptions

# -------------------------------------------------------------------
package OpenILS::EX::Jabber;
use base 'OpenILS::EX::ERROR';
our $ex_msg_header = "Jabber Exception";

package OpenILS::EX::JabberDisconnected;
use base 'OpenILS::EX::ERROR';
our $ex_msg_header = "JabberDisconnected Exception";

=head2 OpenILS::EX::Jabber

Thrown when there is a problem using the Jabber service

=cut

package OpenILS::EX::Transport;
use base 'OpenILS::EX::ERROR';
our $ex_msg_header = "Transport Exception";



# -------------------------------------------------------------------
package OpenILS::EX::InvalidArg;
use base 'OpenILS::EX::ERROR';
our $ex_msg_header = "Invalid Arg Exception";

=head2 OpenILS::EX::InvalidArg

Thrown where an argument to a method was invalid or not provided

=cut


# -------------------------------------------------------------------
package OpenILS::EX::NotADomainObject;
use base 'OpenILS::EX::ERROR';
our $ex_msg_header = "Must be a Domain Object";

=head2 OpenILS::EX::NotADomainObject

Thrown where a OpenILS::DomainObject::oilsScalar or
OpenILS::DomainObject::oilsPair was passed a value that
is not a perl scalar or a OpenILS::DomainObject.

=cut


# -------------------------------------------------------------------
package OpenILS::EX::ArrayOutOfBounds;
use base 'OpenILS::EX::ERROR';
our $ex_msg_header = "Tied array access on a nonexistant index";

=head2 OpenILS::EX::ArrayOutOfBounds

Thrown where a TIEd array (OpenILS::DomainObject::oilsArray) was accessed at
a nonexistant index

=cut



# -------------------------------------------------------------------
package OpenILS::EX::Socket;
use base 'OpenILS::EX::ERROR';
our $ex_msg_header = "Socket Exception";

=head2 OpenILS::EX::Socket

Thrown when there is a network layer exception

=cut



# -------------------------------------------------------------------
package OpenILS::EX::Config;
use base 'OpenILS::EX::PANIC';
our $ex_msg_header = "Config Exception";

=head2 OpenILS::EX::Config

Thrown when a package requires a config option that it cannot retrieve
or the config file itself cannot be loaded

=cut


# -------------------------------------------------------------------
package OpenILS::EX::User;
use base 'OpenILS::EX::ERROR';
our $ex_msg_header = "User Exception";

=head2 OpenILS::EX::User

Thrown when an error occurs due to user identification information

=cut

package OpenILS::EX::Session;
use base 'OpenILS::EX::ERROR';
our $ex_msg_header = "Session Error";


1;
