package OpenILS::Shell;
use strict;
use warnings;
use base 'OpenILS';
use OpenILS::Utils;
use OpenILS::Utils::Config;
use OpenILS::System;
use OpenILS::Utils::JabberUtils;
use IO::Socket::INET;
use OpenILS::Utils::Logger qw(:level);
use OpenILS::EX;

$| = 1;

my $logger = "OpenILS::Utils::Logger";

=head1 Description

Simplified TCP socket which acts as an interface to various internal system calls.  The 
shell listens on 'localhost' on port $config->system->shell_port.  The simplest way to use it is to 
connect via telnet on the local machine and enter 'help' for a list of commands.

This service is loaded at system bootstrap.

=cut

my $socket;

$SIG{INT} = sub { 
	if( $socket ) {
		close( $socket ); 
	} 
	exit;
};

# ------------------------------------------

sub listen {

	my( $class ) = @_;
	my $config = OpenILS::Utils::Config->current;
	my $port = $config->system->shell_port;

	for( my $x = 0; $x != 10; $x++ ) {

		$logger->debug( " * Attempting to connect to shell socket...$x" );

		if( $x > 0 ) {
			$logger->debug( " * Sleeping for 5 seconds before retry" );
			sleep 5;
		}

		# The port should be listed as the $port variable, but for some reason,
		# io::socket is telling me 'no such file or directory' when i load pass
		# in the param that way.  will fix later.
		$socket = new IO::Socket::INET(
			LocalAddr	=> 'localhost',
			LocalPort	=> 4546,
			Proto			=> 'tcp',
			Listen		=> 3,
			) or $logger->debug( "Shell socket creation failed: $!" );

		last if $socket;
	}

	throw OpenILS::EX::Socket( 
			"Could not create shell socket: $!" ) unless $socket;

	$logger->debug( " * System Shell socket is opened and listening" );
	while( my $client = $socket->accept() ) {
		if( OpenILS::Utils::safe_fork() ) { 
			$logger->debug( "Shell spawns a new client" );
			close( $client ); 
		}
		else { 
			process( $client ); 
		}
		$logger->debug( "Shell looping" );
	}

	$logger->debug( "Shell exiting" );
	close( $socket );
}

# ------------------------------------------

sub process {

	my( $sock ) = @_;

	while( 1 ) {

		print $sock "oils# ";
		my $input = <$sock>;
		chomp $input;

		next unless $input =~ /\S/;

		if( $input =~ /exit/i or $input =~ /quit/i ) {
			print $sock "Exiting...\n";
			close( $sock );
			return;
		}

		my @output;

		if( $input =~ /^\!/ ) {
			$input =~ s/^\!//;
			handle_admin( $sock, $input );
			next;
		}

		@output = handle( $input );

		if( ! @output ) {
			print $sock "no results\n";
			next;
		}

		foreach my $line (@output) {
			chomp $line;
			print $sock $line . "\n";
		}
	}
}

# ------------------------------------------

sub handle {

	my( $data ) = @_;

	my @request = split( /\ /, $data );
	my $command = shift @request;

	if( $command =~ /show/i ) {
		return handle_show( @request );
	}

	if( $command =~ /help/i ) {
		return handle_help();
	}

	if( $command =~ /^\// ) {
		my @ret;
		my $stuff = "$command @request";
		$stuff =~ s/^\///;
		return () unless $stuff;
		my @stuff = split( /\ /, $stuff );
		eval{ open( OUT, "@stuff |") };
		if( $@ ) { return (); }
		while( my $blah = <OUT> ) {
			push( @ret, $blah );
		}
		return @ret if( @ret );
		return ("ok");
	}

	return ();
}

# ------------------------------------------

sub handle_show {

	my @request = @_;

	if( $request[0] =~ /jabber_users/i ) {
		return sort OpenILS::Utils::JabberUtils->online_users();
	}

	if( $request[0] =~ /process/i ) {
		my @a;
		my $sys = OpenILS::System->instance();
		foreach my $proc ( keys %{$sys->pid_hash} ) {
			my $pid = $proc;
			my $action = $sys->pid_hash->{$pid};
			push( @a, "$pid : $action" );
		}
		return @a;
	}
	return ();
}

# ------------------------------------------

sub handle_help {

	return ( 
				"-" x 60,
				"exit \t\t\t Exits the shell", 
				"quit \t\t\t Exits the shell",
				"/<command> \t\t Executes a system command (e.g. /ls -l)",
				"show jabber_users \t Shows who's connected Jabber",
				"show process \t\t Shows the currently running system processes",
				#"! killall \t\t Kills all system services including the shell",
				"-" x 60,
			);
}

sub handle_admin {
	my( $sock, $input ) = @_;

	if( $input =~ /killall/i ) {
		print $sock "Are you sure you want to kill all system services?\n";
		print $sock "This will kick you out of the shell.  [yes/no]\n=> ";
		my $response = <$sock>;
		chomp $response;
		if( $response =~ /yes/i ) { 
			print $sock "Killing system services...\n";
			$logger->debug("killall called from system shell");
			# kill parent's parent
			kill( 'INT', _get_parent( getppid() ) );
		}
	}

}

sub _get_parent {
	my $mypid = shift;
	open( PS, "ps -ef|");
	while( <PS> ) {
		chomp;
		my @psdata = split(' ', $_, 8 );
		my $pid = $psdata[1];
		my $ppid = $psdata[2];
		if( $mypid eq $pid ) {
			return $ppid;
		}
	}
}

1;
