package OpenILS::Utils::Log;
use vars qw/$AUTOLOAD @ISA/;
@ISA = qw/OpenILS::Utils/;
use FileHandle;
use Exporter;
use OpenILS::Utils::Config;

our $VERSION = do { my @r=(q$Revision: 1.1 $=~/\d+/g); sprintf "%d."."%02d"x$#r,@r };

{
	my %DBG_LVL = ( ALL	=> -1,
			NONE	=> 0,
			ERROR	=> 1,
			WARNING	=> 3,
			OS	=> 8,
			USER	=> 10,
			INFO	=> 15,
			DEBUG	=> 25,
			INTERNAL=> 64 );

	no strict "refs";
	for my $datum ( keys %DBG_LVL ) {
		*$datum = sub {
			if (lc($_[0]) =~ /^(\d+)$/ || lc($_[1]) =~ /^(\d+)$/) {
				my $number = $1;
				return [map { $DBG_LVL{$_} eq $number ? $_ : () } keys %DBG_LVL]->[0];
			}
			return $DBG_LVL{$datum};
		}
	}
}

%EXPORT_TAGS = (levels => [qw(ALL ERROR WARNING OS USER INFO DEBUG INTERNAL)]);
Exporter::export_ok_tags('levels');

=head2 OpenILS::Utils::Log->debug('message'[, LEVEL])

 Writes B<message> to the debug log if B<LEVEL> is less than
 $Config->logging->debug_level: 

=over 4

=item *

ALL      => -1

=item *

NONE    => 0

=item *

ERROR    => 1

=item *

WARNING  => 3

=item *

OS       => 8

=item *

USER     => 10

=item *

INFO     => 15

=item *

DEBUG    => 25

=item *

INTERNAL => 64

=back

The debug log is specified in this object's config file as:

        [logs]
        debug=debug.log

If no debug log is available B<STDERR> is used.


=cut
sub debug {
        no warnings;
        my $class = shift;
	$class = ref($class) || $class;

        my $msg = shift;

        my $level = shift || $class->INFO;
	my $level_name;
	if ($level =~ /^\d+$/) {
        	$level_name = $class->INFO($level);
	} else {
		$level_name = $level;
		$level = $class->$level_name();
	}

	my $threshold = $class->INFO;
	my $threshold_name = 'INFO';

	if (OpenILS::Utils::Config->current && OpenILS::Utils::Config->current->logging) {
		my $temp_thresh_name = OpenILS::Utils::Config->current->logging->debug_level;
		if ($temp_thresh_name) {
			$threshold_name = uc $temp_thresh_name;
			$threshold = $class->$threshold_name;
		}
	}
        
	return $threshold unless ($msg);
        return unless ($threshold >= $level or $threshold < 0);

        #my ($pkg, $file, $line, $sub) = caller(1);
        my ($pkg, $file, $line, $sub) = caller(0);
	my $debug_msg = $msg;
        $debug_msg =~ s/^/$level_name($0) - $pkg\:\:\[$file\:$line\] - /mg;

	if (OpenILS::Utils::Config->current && OpenILS::Utils::Config->current->logging) {
        	$class->error($msg) if ($level == 1 or $level == 3);
	}

        $class->write_log($debug_msg, 'debug');
}

=head2 OpenILS::Utils::Log->error('message')
 
Writes B<message> to the error log.  The error log is specified
in this object's config file as:
 
        [logs]
        error=error.log
 
If no error log is available B<STDERR> is used.
 
 
=cut

sub error {
        no warnings;
        my $self = shift;
        my $msg = shift;

        my ($pkg, $file, $line, $sub) = caller(1);
        ($pkg, $file, $line, $sub) = caller(0) unless ($pkg);
        $msg =~ s/^/$pkg\:\:\[$file\:$line\] ($0) - /mg;

        $self->write_log($msg, 'error');
}

=head2 OpenILS::Utils::Log->write_log('message', 'logname')
 
Write B<message> to the named log.  The log is specified
in this object's config file as:
 
        [logs]
        testdata=omnibustraps.log
 
If no  log is available B<STDERR> is used.
 
 
=cut

sub write_log {
	my $class = shift;
	$class = ref($class) || $class;

	my $msg = shift;
	my $log = shift;
	$msg =~ s/^/'[' . localtime() . '] '/emg;
	if (OpenILS::Utils::Config->current && OpenILS::Utils::Config->current->logs && $log && OpenILS::Utils::Config->current->logs->$log) {
		my $FH = FileHandle->new('>>'.OpenILS::Utils::Config->current->logs->$log) or goto STD;
		$FH->print($msg . "\n");
		$FH->close();
		return 1;
	}

STD:
	$log = uc($log);
	$msg =~ s/^/$log => /mg;
	warn $msg . "\n";
	return 1;
}

1;

