package OpenILS::Utils::LogServerPrefork;
use strict;
use base qw( OpenILS::Utils Net::Server::PreForkSimple );
use OpenILS::Utils::Logger;
use Time::HiRes qw(gettimeofday);

# --- Run via bash like like so:
# PROTO=$(perl -MOpenILS::Utils::LogServerPrefork -e 'print OpenILS::Utils::LogServerPrefork->PROTO')
# PORT=$(perl -MOpenILS::Utils::LogServerPrefork -e 'print OpenILS::Utils::LogServerPrefork->PORT')
# /usr/bin/perl -MOpenILS::Utils::LogServerPrefork \
#    -e "OpenILS::Utils::LogServerPrefork->run( port => '$PORT/$PROTO' )" &

=head2 Name

OpenILS::Utils::LogServerPrefork

=cut

=head2 Synopsis

 my $port  = OpenILS::Utils::LogServerPrefork->PORT;
 my $proto = OpenILS::Utils::LogServerPrefork->PROTO;

 OpenILS::Utils::LogServerPrefork->run( port => '$port/$proto' );

=cut

=head2 Description

LogServerPrefork is a log server that implements Net::Server's prefork 'personality' for 
handling incoming requests.  It currently listents on udp port 4545, however if module
constants are used judiciously, the actual port and protocol should be transparent.
The server reads the incoming data and routes the log message to the appropriate log file based
on the log route entry in the packet.

=cut

=head2 Protocol

 The protocol is as simple as it gets.  The data packet is formatted like so:

 $log_route|$log_level|$message

 Anything after the second pipe (|) is considered the content of the 
 message to be logged.

=cut


sub LOGGER {"OpenILS::Utils::Logger"}

# --- Create the logs dir if necessary
`echo "if [ ! -d /tmp/oils_logs ]; then mkdir /tmp/oils_logs;fi" | bash`;

# --- Log file locations

sub BASE	{"/tmp/oils_logs/"}
sub ENDING {".log"}

sub L_ERROR			{BASE		. LOGGER->R_ERROR		. ENDING}
sub L_DEBUG			{BASE		. LOGGER->R_DEBUG		. ENDING}
sub L_OPAC			{BASE		. LOGGER->R_OPAC		. ENDING}
sub L_CAT			{BASE		. LOGGER->R_CAT		. ENDING}
sub L_CIRC			{BASE		. LOGGER->R_CIRC		. ENDING}
sub L_REPORTS		{BASE		. LOGGER->R_REPORTS	. ENDING}
sub L_EXSUMMARY	{BASE		. LOGGER->R_EXSUMMARY. ENDING}

# --- Server Constants

sub PORT {'4545'}
sub BUF_SIZE {'1024'}
sub PROTO {"udp"}

my $local_port = PORT;
my $proto = PROTO;

# --- Process Requests

sub process_request {
	my $self = shift;

	my $string;
	$string = $self->{server}->{udp_data};
	if( $string and defined( $string ) ) {
		process_string( $string );
	}
}

sub process_string {
	my $client = shift;
	my @params = split(/\|/,$client);
	my $log = shift @params;
	my $level = shift @params;

	# Put |'s back in since they are stripped 
	# from the message by 'split'
	my $message;
	if( @params > 1 ) {
		foreach my $param (@params) {
			if( $param ne $params[0] ) {
				$message = $message . "|";
			}
			$message = $message . $param;
		}
	}
	else{ $message = "@params"; }


	my $filehandle;
	if( $log eq LOGGER->R_OPAC ) {
		$filehandle = FileHandle->new( '>>' . L_OPAC );
	}
	elsif( $log eq LOGGER->R_EXSUMMARY ) {
		$filehandle = FileHandle->new( '>>' . L_EXSUMMARY );
	}
	elsif( $log eq LOGGER->R_DEBUG ) {
		$filehandle = FileHandle->new( '>>' . L_DEBUG );
	}
	elsif( $log eq LOGGER->R_CIRC ) {
		$filehandle = FileHandle->new( '>>' . L_CIRC );
	}
	elsif( $log eq LOGGER->R_CAT ) {
		$filehandle = FileHandle->new( '>>' . L_CAT );
	}

	elsif( ! $filehandle ) {
		$filehandle = FileHandle->new( '>>' . L_ERROR );
	}

	foreach my $msg ( split(/\n/, $message ) ){
		my $print = format_time() . "|$level|$msg\n";
		$filehandle->print( $print );
	}

	$filehandle->close();
}

sub format_time {
	my ($s, $ms) = gettimeofday();
	my @time = localtime( $s );
	$ms = substr( $ms, 0, 3 );
	my $year = $time[5] + 1900;
	my $mon = $time[4] + 1;
	my $day = $time[3];
	my $hour = $time[2];
	my $min = $time[1];
	my $sec = $time[0];
	$mon = "0" . "$mon" if ( length($mon) == 1 );
	$day = "0" . "$day" if ( length($day) == 1 );
	$hour = "0" . "$hour" if ( length($hour) == 1 );
	$min = "0" . "$min" if (length($min) == 1 );
	$sec = "0" . "$sec" if (length($sec) == 1 );
	return "$year:$mon:$day:$hour:$min:$sec:$ms";
}

1;


